
/*
generic.exe - generic WIN32 application with a status bar
Copyright (C) 2000  Antti Markus (antti@pld.ttu.ee)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef __GNUC__
//useful defines
#define STRICT
#define WIN32_LEAN_AND_MEAN
#endif

//standard includes
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>                                                         

#ifdef __GNUC__
//some defines that were not included in MINGW32 earlier versions
#ifndef HANDLE_WM_LBUTTONDOWN
#define HANDLE_WM_LBUTTONDOWN(h, w, l, fn) \
        __CRACK_ZERO_F(fn, ((h), FALSE, (int)(short)LOWORD(l), (int)(short)HIWORD(l), (UINT)(w)))
#endif

#ifndef HANDLE_WM_MENUSELECT
#define HANDLE_WM_MENUSELECT(hwnd, wParam, lParam, fn) \
    __CRACK_ZERO_F((fn), ((hwnd), (HMENU)(lParam),  \
    (HIWORD(wParam) & MF_POPUP) ? 0L : (WORD)(LOWORD(wParam)), \
    (HIWORD(wParam) & MF_POPUP) ? GetSubMenu((HMENU)lParam, LOWORD(wParam)) : 0L, \
    (UINT)(((short)HIWORD(wParam) == -1) ? 0xFFFFFFFF : HIWORD(wParam))))
#endif

#endif  //ifdef __GNUC__

//additional includes
#include "resource.h"

//*************************************************************
//  FUNCTIONS
//*************************************************************

//entry point for a windows program
int WINAPI WinMain(HINSTANCE, HINSTANCE, LPSTR, int);

//main window callback function
LRESULT CALLBACK WindowFunc(HWND, UINT, WPARAM, LPARAM);

//message handler functions for the main window
BOOL OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct);
void OnPaint(HWND hwnd);
void OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags);
void OnSize(HWND hwnd, UINT state, int cx, int cy);
void OnMenuSelect(HWND hwnd, HMENU hmenu, int item, HMENU hmenuPopup, UINT flags);
void OnCommand(HWND hwnd, WORD id, HWND hwndCtl, UINT codeNotify);
void OnDestroy(HWND hwnd);

//centers child window in respect to parent window
BOOL CenterWnd(const HWND, const HWND);

//about box callback function
BOOL CALLBACK AboutFunc(HWND, UINT, WPARAM, LPARAM);

//*************************************************************

//*************************************************************
//  GLOBAL VARIABLES
//*************************************************************

const char* pClassName="TTWin32Generic";
const int nStatusBarID=200;
HINSTANCE hMainInst=0;  //instance of the application
HWND hMainWnd=0;        //handle to main window of the application
HWND hStatus=0;         //handle to the status bar

//various variables used with virtual window
HDC memdc=0, olddc=0; //device contexts
HBITMAP hbit=0, oldbit=0; //bitmaps
HBRUSH hbrush=0;  //background brush
HPEN hpen=0, hOldPen=0; //pen for drawing
RECT rClient; //window client rectangle
int MaxX=0, MaxY=0; //screen width and height
int nStatusBarH=0;  //status bar height

//*************************************************************

int WINAPI WinMain(HINSTANCE hInstance,
                   HINSTANCE hPrevInstance,
                   LPSTR lpszArgs,
                   int nWinMode)
{

  //let's allow only one instance of the app
  //(code by Robert B. Hess)
  HWND hOldWnd=FindWindow(pClassName, (LPCTSTR)NULL);
  if(hOldWnd)
  {
		// We found another version of ourself. Lets defer to it:
    if(IsIconic(hOldWnd))
    {
      ShowWindow(hOldWnd, SW_RESTORE);
    }
    SetForegroundWindow(hOldWnd);
		// If this app actually had any functionality, we would
		// also want to communicate any action that our 'twin'
		// should now perform based on how the user tried to
		// execute us.
    return FALSE;
  }

  //create a window class
  //by filling in a WNDCLASSEX structure

  WNDCLASSEX WinClass;

  WinClass.cbSize=sizeof(WNDCLASSEX);   //size of the structure
  WinClass.hInstance=hInstance; //instance of the application
  hMainInst=hInstance;
  WinClass.lpszClassName=pClassName;  //window class name
  WinClass.lpfnWndProc=WindowFunc;  //callback function for the application
  //CS_HREDRAW | CS_VREDRAW style causes annoying flicker
  //WinClass.style=CS_HREDRAW | CS_VREDRAW;
  WinClass.style=0;

  //load default application icon
  //WinClass.hIcon=LoadIcon(NULL, IDI_APPLICATION);

  //load standard (32 x 32) icon
  WinClass.hIcon=LoadIcon(hInstance, MAKEINTRESOURCE(IDI_BIG));

  //load small (16 x 16) icon
  //if specified then it is used on the taskbar and in the system menu
  //otherwise 32 x 32 icon is scaled to 16 x 16 by Windows
  //WinClass.hIconSm=LoadIcon(hInstance, MAKEINTRESOURCE(IDI_SMALL));
  WinClass.hIconSm=0;
  //load cursor
  WinClass.hCursor=LoadCursor(NULL, IDC_ARROW);

  //here we can specify a menu handle
  WinClass.lpszMenuName=MAKEINTRESOURCE(IDR_MENU);

  //here we can store some application specific data
  //(we don't use that opportunity right now)
  WinClass.cbClsExtra=0;
  WinClass.cbWndExtra=0;

  //set background to white
  WinClass.hbrBackground=(HBRUSH)GetStockObject(WHITE_BRUSH);

  //register window class we just created
  if(!RegisterClassEx(&WinClass))
    return 0; //registration failed

  //create window
  HWND hwnd=CreateWindow(pClassName,
                         "Generic Windows 95 application",
                         WS_OVERLAPPEDWINDOW,
                         CW_USEDEFAULT,
                         CW_USEDEFAULT,
                         CW_USEDEFAULT,
                         CW_USEDEFAULT,
                         HWND_DESKTOP,
                         NULL,
                         hInstance,
                         NULL);

  if(!hwnd)   //window creation failed
    return 0;

  hMainWnd=hwnd;

  //show window
  ShowWindow(hwnd, nWinMode);
  //force a repaint
  UpdateWindow(hwnd);

  MSG WinMsg;

  //start message pump
  while(GetMessage(&WinMsg, NULL, 0, 0))
  {
    TranslateMessage(&WinMsg);
    DispatchMessage(&WinMsg);
  }

  return WinMsg.wParam;
}

LRESULT CALLBACK WindowFunc(HWND hwnd, UINT Msg,
                            WPARAM wParam, LPARAM lParam)
{
  //device context for drawing
  HDC hdc=0;

  switch(Msg)
  {
    //HANDLE_MSG macro extracts all meaningful information from
    //wParam and lParam, casts them into correct types and gives
    //these values as arguments to handler function
    HANDLE_MSG(hwnd, WM_CREATE, OnCreate);
    HANDLE_MSG(hwnd, WM_PAINT, OnPaint);
    HANDLE_MSG(hwnd, WM_LBUTTONDOWN, OnLButtonDown);
    HANDLE_MSG(hwnd, WM_SIZE, OnSize);
    HANDLE_MSG(hwnd, WM_MENUSELECT, OnMenuSelect);
    HANDLE_MSG(hwnd, WM_COMMAND, OnCommand);
    HANDLE_MSG(hwnd, WM_DESTROY, OnDestroy);

    default:  //all other messages are handled by default
      return DefWindowProc(hwnd, Msg, wParam, lParam);
  }
  return 0;
}

BOOL CALLBACK AboutFunc(HWND hdwnd, UINT Msg, WPARAM wParam, LPARAM lParam)
{

  switch(Msg)
  {
    case WM_INITDIALOG:
      //here we perform all the initialization
      CenterWnd(hdwnd, hMainWnd);
      return TRUE;

    case WM_COMMAND:
    {
      WORD wID=LOWORD(wParam);  //get ID of the button

      switch(wID)
      {
        //not much choice here - quit about box
        case IDOK:
        case IDCANCEL:
          EndDialog(hdwnd, TRUE);
          break;
      }

      return TRUE;
    }

    default:
      return FALSE;   //try changing this line to return TRUE; - neat, huh? :)
  }
  return FALSE;	//just in case...
}

//****************************************************************************
//message handler functions for the main window
//****************************************************************************

BOOL OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
{
  //load common controls DLL
  InitCommonControls();

  //create status bar with the id 200
  hStatus=CreateStatusWindow(WS_CHILD | WS_VISIBLE, (LPCTSTR)NULL, hwnd, nStatusBarID);

  //set it to be a simple status bar (one part only)
  SendMessage(hStatus, SB_SIMPLE, (WPARAM)(BOOL)TRUE, (LPARAM)0);

  RECT rStatus;
  GetWindowRect(hStatus, &rStatus);
  nStatusBarH=rStatus.bottom-rStatus.top;

  //set status bar text
  //SendMessage(hStatus, SB_SETTEXT, (WPARAM)255, (LPARAM)"Ready");

  //get height and width of the screen
  MaxX=GetSystemMetrics(SM_CXSCREEN);
  MaxY=GetSystemMetrics(SM_CYSCREEN);

  //get device context
  HDC hdc=GetDC(hwnd);

  //create compatible device context to memory
  memdc=CreateCompatibleDC(hdc);
  hbit=CreateCompatibleBitmap(hdc, MaxX, MaxY);

  //store old bitmap and select new bitmap into device context
  oldbit=(HBITMAP)SelectObject(memdc, hbit);

  hpen=CreatePen(PS_SOLID, 1, RGB(0, 0, 0));
  hOldPen=(HPEN)SelectObject(memdc, hpen);

  //if we did not set window background
  //while registering window class then
  //we need following two lines
  hbrush=(HBRUSH)GetStockObject(WHITE_BRUSH);
  SelectObject(memdc, hbrush);

  PatBlt(memdc, 0, 0, MaxX, MaxY, PATCOPY);
  ReleaseDC(hwnd, hdc);

  //now we may use memdc instead of hdc in GDI drawing
  //routines (i.e. LineTo(memdc, 0, MaxY);). Afterwards
  //we only need to invalidate the window (or a section of
  //it) by using InvalidateRect. This method's main advantage
  //is avoidance of screen flickering while updating contents
  //of the window
  return TRUE;
}

void OnPaint(HWND hwnd)
{
  //check if there is need to repaint the window
  //in case of internal WM_PAINT messages the update
  //region might be empty
  RECT rUpdate;
  GetUpdateRect(hwnd, &rUpdate, TRUE);
  if(rUpdate.bottom && rUpdate.right)
  {
    //repaint invalid portion of the window
    PAINTSTRUCT PaintStruct;
    HDC hdc=BeginPaint(hwnd, &PaintStruct);
    BitBlt(hdc, 
           rUpdate.left, 
           rUpdate.top, 
           rUpdate.right, 
           rUpdate.bottom, 
           memdc, 
           rUpdate.left, 
           rUpdate.top, SRCCOPY);
    EndPaint(hwnd, &PaintStruct);
  }
}

void OnLButtonDown(HWND hwnd, BOOL fDoubleClick, int x, int y, UINT keyFlags)
{
  //if we press left mouse button...

  //some drawing code...
  RECT rClient;
  GetClientRect(hwnd, &rClient);
  rClient.bottom-=nStatusBarH;

  for(int nCnt=0; nCnt < rClient.right; nCnt+=5)
  {
    MoveToEx(memdc, nCnt, 0, (LPPOINT)NULL);
    LineTo(memdc, rClient.right-nCnt, rClient.bottom);
  }

  //invalidate window to force a WM_PAINT message
  //we are excluding status bar from invalid region
  InvalidateRect(hwnd, &rClient, TRUE);
}

void OnSize(HWND hwnd, UINT state, int cx, int cy)
{
  //tell status bar to size itself
  GetClientRect(hwnd, &rClient);
  SendMessage(hStatus, WM_SIZE, (WPARAM)0, (LPARAM)0);
}

void OnMenuSelect(HWND hwnd, HMENU hmenu, int item, HMENU hmenuPopup, UINT flags)
{
  //item menu item ID
  //if mouse cursor is over menu item
  switch(item)
  {
    //set status bar text according to menu item ID
    case IDM_EXIT:
      SendMessage(hStatus, SB_SETTEXT, (WPARAM)255, (LPARAM)"Exit application");
      break;

    case IDM_HELPTOPICS:
      SendMessage(hStatus, SB_SETTEXT, (WPARAM)255, (LPARAM)"Displays help file");
      break;

    case IDM_ABOUT:
      SendMessage(hStatus, SB_SETTEXT, (WPARAM)255, (LPARAM)"Displays about menu");
      break;

    default:
      SendMessage(hStatus, SB_SETTEXT, (WPARAM)255, (LPARAM)NULL);
  }
}

void OnCommand(HWND hwnd, WORD id, HWND hwndCtl, UINT codeNotify)
{
  //id - menu item ID
  switch(id)
  {
    case IDM_FOPEN:
      //File->Open is disabled for now
      break;

    //Help->Topics
    case IDM_HELPTOPICS:
    {
      //display simple help file
      //hard coding help file name is not such a good idea - but we'll do that now
      BOOL bRes=WinHelp(hwnd, "HELP.HLP", HELP_CONTENTS, (DWORD)0);
      if(!bRes) //if activation of Windows help fails
        MessageBox(GetFocus(),"Unable to activate help", "Generic", MB_OK | MB_ICONHAND);
      break;
    }

    //Help->About
    case IDM_ABOUT:
    {
      //display about box
      int nRes=DialogBox(hMainInst, MAKEINTRESOURCE(IDD_ABOUT), hwnd, (DLGPROC)AboutFunc);
      if(nRes == -1)  //if error displaying about box
      {
        MessageBox(hwnd, "DialogBox failed", "Generic", MB_ICONEXCLAMATION | MB_OK);
      }
      break;
    }
        
    //File->Exit  
    case IDM_EXIT:
      DestroyWindow(hwnd);
      break;         
  }
}

void OnDestroy(HWND hwnd)
{
  //release used resources
  SelectObject(memdc, oldbit);
  SelectObject(memdc, hOldPen);
  DeleteObject(hpen);
  DeleteObject(hbit);
  DeleteDC(memdc);

  //get rid of the help window
  WinHelp(hwnd, "HELP.HLP", HELP_QUIT, (DWORD)0);

  PostQuitMessage(0);
}

//****************************************************************************

//this function centers child window
//in respect to the parent window
//(this function is by Robert B. Hess)

BOOL CenterWnd(const HWND hChild, const HWND hParent)
{
  //get child window information
  RECT rChild;
  GetWindowRect(hChild, &rChild);
  int nChildW=rChild.right-rChild.left;
  int nChildH=rChild.bottom-rChild.top;

  //get parent window information
  RECT rParent;
  GetWindowRect(hParent, &rParent);
  int nParentW=rParent.right-rParent.left;
  int nParentH=rParent.bottom-rParent.top;

  // Get the limits of the 'desktop'
  RECT rDesktop;
	BOOL bRes=SystemParametersInfo(SPI_GETWORKAREA, sizeof(RECT), &rDesktop, 0);
	if(!bRes) //if we did not get limits to the 'desktop'
  {
		rDesktop.left=rDesktop.top = 0;

    //if we did not have global variables
    //storing width and height of the screen
    //then we would need to uncomment following two rows
		//rDesktop.right=GetSystemMetrics(SM_CXSCREEN);
		//rDesktop.bottom=GetSystemMetrics(SM_CYSCREEN);

    //let's use our previously found values
    rDesktop.right=MaxX;
    rDesktop.bottom=MaxY;
	}

  //calculate new coordinates for the child window
  int nNewX=rParent.left+((nParentW-nChildW) >> 1); //x >> 1 == x/2
  if(nNewX < rDesktop.left)
    nNewX=rDesktop.left;
  else if((nNewX+nChildW) > rDesktop.right)
    nNewX=rDesktop.right-nChildW;

  int nNewY=rParent.top+((nParentH-nChildH) >> 1); //x >> 1 == x/2
  if(nNewY < rDesktop.top)
    nNewY=rDesktop.top;
  else if((nNewY+nChildH) > rDesktop.bottom)
    nNewY=rDesktop.bottom-nChildH;

  //put child window into its proper place
  //retaining window size (SWP_NOSIZE) and current Z order (SWP_NOZORDER)
  return SetWindowPos(hChild, NULL, nNewX, nNewY, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}
